// mat.H Version 4.3  Copleft (c) 杉本典夫
// Time-stamp: <2023-10-25 14:00:26 tac>
//
// 行列クラスmatとベクトルクラスvec用のヘッダファイル
//
// 関連ライブラリ:なし

#ifndef _MAT                    // 重複インクルードを防ぐため
#define _MAT

#include <fun.H>                // 共通関数用ヘッダファイル
//#include <eigen3/Eigen/Eigen>   // 行列ライブラリ用ヘッダファイル(将来のため)

namespace matvec                // 共通関数用名前空間matvec定義
{
  using namespace fun; // 名前空間funを合成する(名前空間stdとboost含む)

  template <class T> class mat  // 行列クラステンプレート
  {
  protected:
    size_t vr, vc;              // vr:行数,vc:列数
    vector<T> dvec;             // 数値ベクトル
  public:
    mat() : vr(0), vc(0), dvec() // デフォールトコンストラクタ
    {}
    explicit mat(size_t n0) : vr(n0), vc(n0), dvec(n0 * n0, (T)0.0)
    // コンストラクタ(正方行列)
    {}
    mat(size_t r0, size_t c0) : vr(r0), vc(c0), dvec(r0 * c0, (T)0.0)
    // コンストラクタ(一般行列)
    {}
    mat(const mat<T>& mat0) : vr(mat0.vr), vc(mat0.vc), dvec(mat0.dvec)
    // コピーコンストラクタ
    {}
    virtual ~mat() // デストラクタ
    {}
    const mat<T>& operator=(const mat<T>& mat0) // コピー代入演算子
    {
      if (&mat0 == this) return *this; // 自分自身への代入を防ぐ
      vr = mat0.vr;  vc = mat0.vc;  dvec = mat0.dvec;
      return *this;             // 自分自身への参照を返す
    }
    const vector<T>& dvector() const // ベクトル配列に対する参照を返す
    { return dvec; }
    const mat<T>& operator=(T a)  // スカラーの代入演算子(全要素同値代入)
    {
      dvec.assign(dvec.size(), a);
      return *this;             // 自分自身への参照を返す
    }
    const mat<T>& operator+=(const mat<T>& mat0) // 行列の加算演算子
    {
      if (vr != mat0.vr || vc != mat0.vc) throw runtime_error("行列の次数が違うもんで加算できせん！");
      for (size_t i = 0; i < dvec.size(); ++i) dvec[i] = dvec[i] + mat0.dvec[i];
      return *this;             // 自分自身への参照を返す
    }
    const mat<T>& operator-=(const mat<T>& mat0) // 行列の減算演算子
    {
      if (vr != mat0.vr || vc != mat0.vc) throw runtime_error("行列の次数が違うもんで減算できせん！");
      for (size_t i = 0; i < dvec.size(); ++i) dvec[i] = dvec[i] - mat0.dvec[i];
      return *this;             // 自分自身への参照を返す
    }
    const mat<T>& operator*=(T a) // スカラーと行列の積演算子(行列のスカラー倍)
    {
      for (size_t i = 0; i < dvec.size(); ++i) dvec[i] = dvec[i] * a;
      return *this;             // 自分自身への参照を返す
    }
    T operator()(size_t ridx, size_t cidx) const // 行列要素指定用演算子
    { return (ridx < vr && cidx < vc) ? dvec[ridx * vc + cidx] : (T)0.0; }
    T maxabse() const           // 絶対値最大要素の絶対値を返す
    {
      T maxe = (T)0.0;
      for (size_t i = 0; i < dvec.size(); ++i)
      {
        T abse = fabs(dvec[i]);
        if (abse > maxe) maxe = abse;
      }
      return maxe;
    }
    const mat<T>& clear()       // 行列の要素を全て削除する
    {
      vr = vc = 0;  dvec.clear();
      return *this;             // 自分自身への参照を返す
    }
    const mat<T>& clean(T eps = (T)0.0) // 行列の要素のうちeps以下のものを0にする
    {
      if (eps <= (T)0.0) eps = numeric_limits<T>::epsilon() * fmax((T)5.0, this->maxabse());
      for (size_t i = 0; i < dvec.size(); ++i)
      {
        if (fabs(dvec[i]) <= eps) dvec[i] = (T)0.0;
      }
      return *this;             // 自分自身への参照を返す
    }
    size_t size() const         // 行列の要素数を返す
    { return dvec.size(); }
    bool empty() const          // 行列の要素が空かどうかを返す
    { return dvec.empty(); }
    size_t r() const            // 行列の行数を返す
    { return vr; }
    size_t c() const            // 行列の列数を返す
    { return vc; }
    T at(size_t ridx, size_t cidx) const // 行列の要素を返す
    {
      if (ridx >= vr || cidx >= vc) throw runtime_error("行列のat添字が範囲外だでかんわ！");
      return dvec[ridx * vc + cidx];
    }
    T assign(size_t ridx, size_t cidx, T a) // 行列の要素を設定する
    {
      if (ridx >= vr || cidx >= vc) throw runtime_error("行列のassign添字が範囲外だでかんわ！");
      return (dvec[ridx * vc + cidx] = a);
    }
    T add(size_t ridx, size_t cidx, T a) // 行列の要素に加算する
    {
      if (ridx >= vr || cidx >= vc) throw runtime_error("行列のadd添字が範囲外だでかんわ！");
      return (dvec[ridx * vc + cidx] = dvec[ridx * vc + cidx] + a);
    }
    const mat<T>& resize(size_t r0, size_t c0) // 行列の次数を変更する
    {
      if (r0 == vr && c0 == vc) return *this; // 無変更
      vector<T> tvec(r0 * c0, (T)0.0);
      size_t rmin = min(r0, vr), cmin = min(c0, vc);
      for (size_t i = 0; i < rmin; ++i)
      {
        size_t toffset = i * c0, roffset = i * vc;
        for (size_t j = 0; j < cmin; ++j) tvec[toffset +  j] = dvec[roffset + j];
      }
      vr = r0;  vc = c0;  dvec = tvec;
      return *this;             // 自分自身への参照を返す
    }
    const mat<T> sub(size_t rsidx, size_t csidx, size_t reidx, size_t ceidx) const
    // 部分行列を返す
    {
      if (rsidx > reidx || reidx >= vr || csidx > ceidx || ceidx >= vc) throw runtime_error("部分行列の範囲指定が不適だでかんわ！");
      mat<T> tmat(reidx - rsidx + 1, ceidx - csidx + 1);
      for (size_t i = 0; i < tmat.r(); ++i)
      {
        size_t toffset = i * tmat.c(), roffset = (rsidx + i) * vc;
        for (size_t j = 0; j < tmat.c(); ++j) tmat.assign(i, j, dvec[roffset + csidx + j]);
      }
      return tmat;              // 部分行列を返す
    }
    virtual const mat<T>& trn() // 行列を転置する(転置行列化)
    {
      vector<T> tvec(vr * vc);
      for (size_t i = 0; i < vc; ++i)
      {
        size_t toffset = i * vr;
        for (size_t j = 0; j < vr; ++j) tvec[toffset + j] = dvec[j * vc + i];
      }
      swap(vr, vc);
      dvec = tvec;
      return *this;             // 自分自身への参照を返す
    }
    const mat<T>& sweep(size_t pidx) // 行列を指定された軸で掃き出す
    {
      if (vr != vc) throw runtime_error("正方行列でないと掃き出し計算はできんよ！");
      if (pidx >= vr) throw runtime_error("行列掃き出し軸の指定が間違っとるよ！");
      size_t poffset = pidx * vc;
      T pivot = dvec[poffset + pidx];
      const T PIVeps = numeric_limits<T>::epsilon() * fmax((T)1.0, this->maxabse());
      if (fabs(pivot) < PIVeps) // 軸の要素が計算誤差以内の時
      {
        dvec[poffset + pidx] = (T)0.0;
        return *this;
      }
      dvec[poffset + pidx] = (T)1.0;
      for (size_t j = 0; j < vc; ++j) dvec[poffset + j] = dvec[poffset + j] / pivot;
      for (size_t i = 0; i < vr; ++i) // r行をc列まで掃き出す
      {
        if (i != pidx)
        {
          size_t roffset = i * vc;
          pivot = dvec[roffset + pidx];
          dvec[roffset + pidx] = (T)0.0;
          for (size_t j = 0; j < vc; ++j) dvec[roffset + j] = dvec[roffset + j] - pivot * dvec[poffset + j];
        }
      }
      return *this;             // 自分自身への参照を返す
    }
    const mat<T>& inv()         // 逆行列にする
    {
      if (vr != vc) throw runtime_error("正方行列でないと逆行列は計算できんよ！");
      return inv(vr);
    }
    const mat<T>& inv(size_t n) // 指定された次数まで逆行列にする
    {
      if (vr != vc) throw runtime_error("正方行列でないと逆行列は計算できんよ！");
      if (n > vr) throw runtime_error("逆行列の次数指定が間違っとるよ！");
      for (size_t i = 0; i < n; ++i) sweep(i);
      return *this;
    }
    T det() const               // 行列式を返す
    {
      if (vr != vc) return (T)0.0; // 非正方行列
      if (vr == 1) return dvec[0];
      if (vr == 2) return dvec[0] * dvec[3] - dvec[1] * dvec[2];
      if (vr == 3)
      {
        T de = dvec[0] * dvec[4] * dvec[8] + dvec[1] * dvec[5] * dvec[6] + dvec[2] * dvec[3] * dvec[7];
        return de - dvec[2] * dvec[4] * dvec[6] - dvec[1] * dvec[3] * dvec[8] - dvec[0] * dvec[5] * dvec[7];
      }
      mat<T> tmat(*this);
      T de = (T)1.0;
      const T Deps = numeric_limits<T>::epsilon() * fmax((T)1.0, this->maxabse());
      for (size_t k = 0; k < tmat.r(); ++k)
      {
        T tkk = tmat(k, k);
        de = de * tkk;
        if (fabs(de) < Deps)    // 行列式の値が計算誤差以内の時
        {
          de = (T)0.0;
          break;
        }
        for (size_t j = k + 1; j < tmat.r(); ++j) tmat.assign(k, j, tmat(k, j) / tkk);
        for (size_t i = k + 1; i < tmat.r(); ++i)
        {
          T tik = tmat(i, k);
          for (size_t j = k + 1; j < tmat.r(); ++j) tmat.assign(i, j, tmat(i, j) - tik * tmat(k, j));
        }
      }
      return de;
    }
    int cholesky(mat<T>& lmat) const
    // 対称行列のコレスキー分解を行い、下三角行列を返す
    {
      if (vr != vc || lmat.r() != lmat.c() || lmat.r() != vr) return disperr("cholesky()", "コレスキー分解ができるのは対称な正方行列のみ！");
      const T Eeps = numeric_limits<T>::epsilon() * fmax((T)1.0, this->maxabse());
      lmat = (T)0.0;
      for (size_t i = 0; i < vr; ++i)
      {
        for (size_t j = 0; j < i; ++j)
        {
          T tmp = (*this)(i, j);
          for (size_t k = 0; k < j; ++k) tmp = tmp - lmat(i, k) * lmat(j, k);
          lmat.assign(i, j, (fabs(lmat(j, j)) < Eeps) ? (T)0.0 : tmp / lmat(j, j));
        }
        T tmp = (*this)(i, i);
        for (size_t k = 0; k < i; ++k) tmp = tmp - lmat(i, k) * lmat(i, k);
        if (tmp < (T)0.0) return disperr("cholesky()", "コレスキー分解不成功！");
        lmat.assign(i, i, sqrt(tmp));
      }
      return 0;                 // 分解成功
    }
  };

  template <class T> class vec : public mat<T> // ベクトルクラス
  {
  public:
    vec() : mat<T>()            // デフォールトコンストラクタ
    {}
    explicit vec(size_t n0) : mat<T>(n0, 1) // コンストラクタ(列ベクトル)
    {}
    vec(const vec<T>& vec0) : mat<T>(vec0) // コピーコンストラクタ
    {}
    explicit vec(const vector<T>& vector0)
    // ベクトル配列によるコンストラクタ(列ベクトル)
    { this->vr = vector0.size();  this->vc = 1;  this->dvec = vector0; }
    vec(const mat<T>& mat0)
    // 行列によるコンストラクタ(行または列ベクトル作成)
    {
      if (mat0.r() > 1 && mat0.c() > 1) // ベクトルではない時
      { this->vr = mat0.size();  this->vc = 1; } // 全要素を列ベクトル化
      else                                               // 行または列ベクトル
      { this->vr = mat0.r();  this->vc = mat0.c(); }
      this->dvec = mat0.dvector();
    }
    ~vec()                      // デストラクタ
    {}
    const vec<T>& operator=(const vec<T>& vec0) // コピー代入演算子
    {
      if (&vec0 == this) return *this; // 自分自身への代入を防ぐ
      this->vr = vec0.vr;  this->vc = vec0.vc;  this->dvec = vec0.dvec;
      return *this;             // 自分自身への参照を返す
    }
    const vec<T>& operator=(T x)  // スカラーの代入演算子(全要素同値代入)
    {
      this->dvec.assign(this->dvec.size(), x);
      return *this;             // 自分自身への参照を返す
    }
    T operator()(size_t idx) const // 要素指定用演算子
    { return (idx < this->dvec.size()) ? this->dvec[idx] : (T)0.0; }
    size_t dim() const          // ベクトルの次元数を返す
    { return this->dvec.size(); }
    T at(size_t idx) const      // ベクトルの要素を返す
    { return (idx < this->dvec.size()) ? this->dvec[idx] : (T)0.0; }
    T assign(size_t idx, T x)   // ベクトルの要素を設定する
    {
      if (idx >= this->dvec.size()) throw runtime_error("ベクトルのassign添字が範囲外だでかんわ！");
      return (this->dvec[idx] = x);
    }
    T add(size_t idx, T x)      // ベクトルの要素を加算する
    {
      if (idx >= this->dvec.size()) throw runtime_error("ベクトルのadd添字が範囲外だでかんわ！");
      return (this->dvec[idx] = this->dvec[idx] + x);
    }
    const vec<T>& resize(size_t n0) // ベクトルの次数を変更する
    {
      if (this->dvec.size() == n0) return *this; // 無変更
      vector<T> tvec(n0, (T)0.0);
      size_t nmin = min(n0, this->dvec.size());
      for (size_t i = 0; i < nmin; ++i) tvec[i] = this->dvec[i];
      if (this->vc == 1) this->vr = n0; // 列ベクトル
      else if (this->vr == 1) this->vc = n0; // 行ベクトル
      else { this->vr = n0;  this->vc = 1; } // 列ベクトルにする
      this->dvec = tvec;
      return *this;             // 自分自身への参照を返す
    }
    const vec<T>& sub(size_t sidx, size_t eidx) const // 部分ベクトルを返す
    {
      if (sidx > eidx || eidx >= this->vr * this->vc) throw runtime_error("部分ベクトルの範囲指定が不適だでかんわ！");
      size_t n = eidx - sidx + 1;
      static vec<T> xv(n);
      if (this->vc > 1) xv.trn(); // 行ベクトルにする
      for (size_t i = 0; i < n; ++i) xv.assign(i, this->dvec[sidx + i]);
      return xv;                // 部分ベクトルへの参照を返す
    }
    const vec& trn()            // 行ベクトル←→列ベクトル変換
    {
      swap(this->vr, this->vc);
      return *this;             // 自分自身への参照を返す
    }
    T norm() const              // ベクトルの大きさ(ノルム)を返す
    {
      T nm = (T)0.0;
      for (size_t i = 0; i < this->dvec.size(); ++i) nm = nm + this->dvec[i] * this->dvec[i];
      const T NMeps = numeric_limits<T>::epsilon() * fmax((T)1.0, this->maxabse());
      if (fabs(nm) < NMeps) return (T)0.0;
      return sqrt(nm);
    }
  };

  template <class T>
  const mat<T> operator+(const mat<T>& amat, const mat<T>& bmat) // 行列同士の和
  {
    if (amat.r() != bmat.r() || amat.c() != bmat.c()) throw runtime_error("2つの行列の次数が違うもんで和が計算できせん！");
     mat<T> cmat(amat);         // 和を代入する行列
    cmat += bmat;
    return cmat;                // 和行列を返す
  }

  template <class T>
  const mat<T> operator-(const mat<T>& amat, const mat<T>& bmat) // 行列同士の差
  {
    if (amat.r() != bmat.r() || amat.c() != bmat.c()) throw runtime_error("2つの行列の次数が違うもんで差が計算できせん！");
    mat cmat(amat);             // 差を代入する行列
    cmat -= bmat;
    return cmat;                // 差行列を返す
  }

  template <class T>
  const mat<T> operator*(const mat<T>& amat, const mat<T>& bmat) // 行列同士の積
  {
    if (amat.c() != bmat.r()) throw runtime_error("2つの行列の列数と行数が違うもんで積が計算できせん！");
    mat<T> cmat(amat.r(), bmat.c()); // 積を代入する行列
    for (size_t i = 0; i < amat.r(); ++i)
    {
      for (size_t j = 0; j < bmat.c(); ++j)
      {
        T eij = (T)0.0;
        for (size_t k = 0; k < amat.c(); ++k) eij = eij + amat(i, k) * bmat(k, j);
        cmat.assign(i, j, eij);
      }
    }
    return cmat;                // 積行列を返す
  }

  template <class T>
  bool operator==(const mat<T>& amat, const mat<T>& bmat) // 行列の一致
  {
    if (amat.r() != bmat.r() || amat.c() != bmat.c()) return false; // 次数違いは不一致！
    bool eq = true;
    for (size_t i = 0; eq && i < amat.r(); ++i)
    {
      for (size_t j = 0; eq && j < amat.c(); ++j)
      {
        if (sgndiff(amat(i, j), bmat(i, j)) != 0) eq = false;
      }
    }
    return eq;                  // 一致:ture,不一致:falseを返す
  }

  template <class T>
  bool operator!=(const mat<T>& amat, const mat<T>& bmat) // 行列の不一致
  { return !(amat == bmat); }

  template <class T>
  T inp(const vec<T>& xv, const vec<T>& yv) // ベクトルの内積を返す
  {
    if (xv.dim() != yv.dim()) throw runtime_error("ベクトルの次元数が違うもんで内積が計算できせん！");
    T inp = (T)0.0;
    for (size_t i = 0; i < xv.dim(); ++i) inp = inp + xv(i) * yv(i);
    const T INPeps = numeric_limits<T>::epsilon() * fmax((T)1.0, fmax(xv.maxabse(), yv.maxabse()));
    if (fabs(inp) < INPeps) return (T)0.0;
    return inp;
  }

  template <class T>
  T eigen(const mat<T>& amat, vec<T>& xv)
  // 対称行列の固有値と固有ベクトルを返す(ベキ乗法)
  {
    if (amat.r() < 1 || amat.r() != amat.c()) throw runtime_error("固有値は正方行列しか計算できんよ！");
    if (amat.r() != xv.r()) throw runtime_error("行列とベクトルの次数が違うので固有値が計算できんよ！");
    xv = (T)0.9;
    xv.assign(0, (T)1.0); // x == 0.9 だとPANOVA計算上不都合があるため
    vec<T> xov(xv);
    T eval = (T)0.0, eold = (T)0.0;
    do
    {
      xov = xv;  eold = eval;
      xv = amat * xov;
      eval = xv.norm();
      if (eval != (T)0.0) xv *= (T)1.0 / eval;
      else { eold = (T)0.0;  xv = (T)0.0; } // 固有値が誤差以内(計算終了)
    } while (sgndiff(eval, eold) != 0);
    return eval;
  }

  template <class T>
  T eigen(const mat<T>& amat, vec<T>& xv, vec<T>& yv)
  // 非対称行列の固有値と固有ベクトルを返す(ベキ乗法)
  {
    if (amat.r() < 1 || amat.r() != amat.c()) throw runtime_error("固有値は正方行列しか計算できんよ！");
    if (amat.r() != xv.r() || amat.r() != yv.r()) throw runtime_error("行列とベクトルの次数が違うので固有値が計算できんよ！");
    T eval = eigen(amat, xv);
    mat<T> atmat(amat);
    eigen(atmat.trn(), yv);
    return eval;
  }

  template <class T>
  T eigenud(mat<T>& amat, vec<T>& xv)
  // 対称行列の固有値・固有ベクトルの計算と行列の更新
  {
    if (amat.r() < 1 || amat.r() != amat.c()) throw runtime_error("固有値は正方行列しか計算できんよ！");
    if (amat.r() != xv.r()) throw runtime_error("行列とベクトルの次数が違うので固有値が計算できんよ！");
    T eval = eigen(amat, xv);
    if (eval == (T)0.0) return eval;
    vec<T> xtv(xv);
    T inpxx = inp(xv, xv);
    if (inpxx != (T)0.0) xtv *= eval / inpxx;
    mat<T> xxmat(xv * xtv.trn());
    amat -= xxmat;
    amat.clean();
    return eval;
  }

  template <class T>
  T eigenud(mat<T>& amat, vec<T>& xv, vec<T>& yv)
  // 非対称行列の固有値・固有ベクトルの計算と行列の更新
  {
    if (amat.r() < 1 || amat.r() != amat.c()) throw runtime_error("固有値は正方行列しか計算できんよ！");
    if (amat.r() != xv.r() || amat.r() != yv.r()) throw runtime_error("行列とベクトルの次数が違うので固有値が計算できんよ！");
    T eval = eigen(amat, xv, yv);
    if (eval == (T)0.0) return eval;
    T inpxy = inp(xv, yv);
    if (inpxy == (T)0.0) inpxy = xv.norm();
    vec<T> ytv(yv);
    if (inpxy != (T)0.0) ytv *= eval / inpxy;
    mat<T> xymat(xv * ytv.trn());
    amat -= xymat;
    amat.clean();
    return eval;
  }

} // 名前空間matvec終わり

#endif
